"""Module for Beryllium Lens Calculations."""
import json
import logging
import os
import shutil
from datetime import date
from itertools import chain, product

import numpy as np

from .constants import chemical_name_to_formula, density

logger = logging.getLogger(__name__)
# global variable to help with testing the plan_set function
_plan_set_test_res = None

# Constant for converting between FWHM and sigma of a Gaussian function.
FWHM_SIGMA_CONVERSION = 2.35482004503
# Constant for converting between wavelength and photon energy.
WAVELENGTH_PHOTON = 1.2398
# Path the the lens_set file. Users shuld use :meth:`configure_lens_set_file`
# to configure it to the correct path
LENS_SET_FILE = None

# full width at half maximum unfocused (in meters)
FWHM_UNFOCUSED = 500e-6
# Disk Thickness in meters
DISK_THICKNESS = 1.0e-3
# Apex of the lens in meters
APEX_DISTANCE = 30e-6
# Distance from the lenses to the sample (in meters)
DISTANCE = 4.0
# Atomic symbol for element, defaults to 'Be'
MATERIAL = 'Be'
# Set of Be lenses with thicknesses (thicknesses in meters)
LENS_RADII = [50e-6, 100e-6, 200e-6, 300e-6,
              500e-6, 1000e-6, 1500e-6, 2000e-6, 3000e-6]


def configure_defaults(fwhm_unfocused=None, disk_thickness=None,
                       apex_distance=None, distance=None,
                       material=None, lens_radii=None):
    """
    Configure defaults.

    Parameters
    -----------
    fwhm_unfocused : float, optional
        Full width at half maximum unfocused in meters.
    disk_thickness : float, optional
        Disk thickness in meters.
    apex_distance : float, optional
        Apex Distance in meters.
    distance : float, optional
        Distance from the lenses to the sample in meters.
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    lens_radii : list
         Set of Be lenses with thicknesses in meters.

    Examples
    --------
    >>> configure_defaults(distance=3.852)
    """
    global FWHM_UNFOCUSED
    global DISK_THICKNESS
    global APEX_DISTANCE
    global DISTANCE
    global MATERIAL
    global LENS_RADII

    FWHM_UNFOCUSED = fwhm_unfocused or FWHM_UNFOCUSED
    DISK_THICKNESS = disk_thickness or DISK_THICKNESS
    APEX_DISTANCE = apex_distance or APEX_DISTANCE
    DISTANCE = distance or DISTANCE
    MATERIAL = material or MATERIAL
    LENS_RADII = lens_radii or LENS_RADII


def configure_lens_set_file(lens_file_path):
    """
    Configure the path to the lens set file.

    Parameters
    ----------
    lens_file_path : str
        Path to the lens_set file.
    """
    global LENS_SET_FILE
    if not os.path.exists(lens_file_path):
        err_msg = f'Provided invalid path for lens set file: {lens_file_path}'
        logger.error(err_msg)
        raise FileNotFoundError(err_msg)
    LENS_SET_FILE = os.path.abspath(lens_file_path)
    return LENS_SET_FILE


def photon_to_wavelength(energy):
    """
    Find the wavelength in micrometers.

    Use photon energy in electronvolts. The equation is approximately: λ[µm] =
    1.2398 / E[eV]. The photon energy at 1 μm wavelength, the wavelength of
    near infrared radiation, is approximately 1.2398 eV.

    Parameters
    ----------
    energy : number
        Photon energy in electronvolts

    Returns
    -------
    Wavelength in micrometers

    Examples
    --------
    >>> photon_to_wavelength(8)
    0.154975
    """
    return WAVELENGTH_PHOTON / energy


def gaussian_sigma_to_fwhm(sigma):
    """
    Convert between FWHM and sigma of a Gaussian function.

    FWHM = 2.35482004503 * sigma

    https://brainder.org/2011/08/20/gaussian-kernels-convert-fwhm-to-sigma/

    Parameters
    ----------
    sigma : float

    Returns
    -------
    FWHM -  Full Width at the Half Maximum

    Examples
    --------
    >>> gaussian_sigma_to_fwhm(0.3)
    0.7064460135089999
    """
    return FWHM_SIGMA_CONVERSION * sigma


def gaussian_fwhm_to_sigma(fwhm):
    """
    Convert between FWHM and sigma of a Gaussian function.

    sigma = FWHM / 2.35482004503

    https://brainder.org/2011/08/20/gaussian-kernels-convert-fwhm-to-sigma/

    Parameters
    ----------
    fwhm : float
        Full Width at the Half Maximum.

    Returns
    -------
    sigma

    Examples
    --------
    >>> gaussian_fwhm_to_sigma(0.023)
    0.009767200703316157
    """
    return fwhm / FWHM_SIGMA_CONVERSION


def get_lens_set(set_number_top_to_bot, filename=None, get_all=False):
    """
    Get the lens set from the file provided.

    Parameters
    ----------
    set_number_top_to_bot : int
        The Be lens holders can take 3 different sets that we usually set
        before experiments, this is to specify what number set.
    filename : str, optional
        File path of the lens_set file.
    get_all : bool
        Indicates if it should return the entire stack of lens.

    Returns
    -------
    lens_set : list
        [numer1, lensthick1, number2, lensthick2 ...]
    """
    if filename is None and LENS_SET_FILE is None:
        err_msg = ('You must provide the path to the lens_set file or you '
                   'must configure it via configure_lens_set_file function')
        logger.error(err_msg)
        raise ValueError(err_msg)
    elif filename is None:
        filename = LENS_SET_FILE
    if not os.path.exists(filename):
        err_msg = f'Provided invalid path for lens set file: {filename}'
        logger.error(err_msg)
        raise FileNotFoundError(err_msg)
    if os.stat(filename).st_size == 0:
        err_msg = (f'The file is empty: {filename}, use set_lens_set_to_file '
                   'to write set lenses to the file.')
        logger.error(err_msg)
        raise ValueError(err_msg)
    with open(filename) as lens_file:
        try:
            sets = json.loads(lens_file.read())
        except json.decoder.JSONDecodeError as err:
            logger.error('When getting the lens set: %s', err)
            raise err

        if get_all:
            return sets

        if set_number_top_to_bot not in range(1, len(sets)):
            err_msg = ('Provided an invalid set_number_top_to_bottom: '
                       f'{set_number_top_to_bot}, please provide a number '
                       f'from 1 to {len(sets)}')
            logger.error(err_msg)
            raise ValueError(err_msg)
    # if only one set in the list, return the list
    if not isinstance(sets[0], list):
        return sets
    else:
        return sets[set_number_top_to_bot - 1]


def set_lens_set_to_file(list_of_sets, filename=None,
                         make_backup=True):
    """
    Write lens set to a file.

    The Be lens holders can take 3 different sets that we usually set before
    few experiments so we only vent the relevant beamline section once. We then
    store these sets into a config file that we can call from some methods.
    Later, we save this config file to the specific experiment so users can
    make sure that they know which stack was used for there beamtime.

    Parameters
    ----------
    list_of_sets : list
        List of lists with lens sets.
    filename : str, optional
        Path to the filename to set the lens sets list to.
        This should be a .npy format file.
    make_backup : bool, optional
        To indicate if a backup file should be created or not.
        Defaults to `True`.

    Examples
    --------
    >>> list_of_sets = [[3, 0.0001, 1, 0.0002],
                        [1, 0.0001, 1, 0.0003, 1, 0.0005],
                        [2, 0.0001, 1, 0.0005]]
    >>> set_lens_set_to_file(list_of_sets, '../path/to/lens_set')
    """
    if filename is None and LENS_SET_FILE is None:
        logger.error('You must provide the path to the lens_set file or you '
                     'must configure it via :meth: `configure_lens_set_file`')
        return
    elif filename is None:
        filename = LENS_SET_FILE
    # Make a backup with today's date.
    if make_backup:
        backup_path = filename + str(date.today()) + '.bak'
        try:
            shutil.copyfile(filename, backup_path)
        except Exception as ex:
            logger.error('Something went wrong with copying the file %s', ex)
            pass
    with open(filename, 'w') as lens_file:
        try:
            lens_file.write(json.dumps(list_of_sets))
        except json.decoder.JSONDecodeError as err:
            logger.error('Something went wrong when writing lens set to the '
                         'file %s', err)
            raise err


def get_att_len(energy, material=None, density=None):
    """
    Get the attenuation length (in meter) of a material.

    The X-ray beam intensity I(x) at depth x in a material is a function of the
    attenuation coefficient mu, and can be calculated by the Beer-Lambert law.
    The Absorption Length (or Attenuation Length) is defined as the distance
    into a material where the x-ray beam intensity has decreased to a value of
    1/e (~ 40%) of the incident beam intensity (Io).

    (1/e) = e^(-mu * x)
    ln(1/e) = ln(e^(-mu * x))
    1 = mu * x
    x = 1/mu

    Parameters
    ----------
    energy : number
        Beam Energy given in KeV
    material : str, optional
        Atomic symbol for element, defaults to 'Be'
    density : float, optional
        Material density in g/cm^3

    Returns
    -------
    att_len : float
        Attenuation length (in meters)

    Raises
    ------
    ValueError
        If an invalid symbol is provided for material.

    Examples
    --------
    >>> get_att_len(energy=8, material='Be')
    0.004810113254120656
    """
    import xraydb as xdb
    material = material or MATERIAL
    density = density or xdb.atomic_density(material)
    try:
        # xdb.material_my returns absorption length in 1/cm and takes energy
        # or array of energies in eV.
        att_len = 1.0 / (xdb.material_mu(material, energy * 1.0e3,
                                         density=density)) * 1.0e-2
    except Exception as ex:
        logger.error('Get Attenuation Length error: %s', ex)
        raise ex
    return att_len


def get_delta(energy, material=None, density=None):
    """
    Calculate delta for a given material at a given energy.

    Anomalous components of the index of refraction for a material, using the
    tabulated scattering components from Chantler.

    Parameters
    ----------
    energy : number
        x-ray energy in KeV
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3

    Returns
    -------
    delta : float
        Real part of index of refraction

    Raises
    ------
    ValueError
        If an invalid symbol is provided for material.
    ZeroDivisionError
        When energy is 0.

    Examples
    --------
    >>> get_delta(energy=8, material='Au')
    4.728879989419882e-05
    """
    import xraydb as xdb
    material = material or MATERIAL
    # xray_delta_beta returns (delta, beta, atlen), wehre delta : real part of
    # index of refraction, and takes x-ray energy in eV.
    if density is None:
        density = xdb.atomic_density(material)
    try:
        delta = xdb.xray_delta_beta(material,
                                    density=density,
                                    energy=energy * 1.0e3)[0]
    except Exception as ex:
        logger.error('Get Delta error: %s', ex)
        raise ex
    return delta


def calc_focal_length_for_single_lens(energy, radius, material=None,
                                      density=None):
    """
    Calculate the Focal Length for a single lens.

    Parameters
    ----------
    energy : number
        Beam Energy in in KeV
    radius : float
        Radius of curvature (in meters)
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3

    Returns
    -------
    focal_length : float
        The focal length for a single lens

    Examples
    --------
    >>> calc_focal_length_for_single_lens(8, 0.03, 'Be')
    2814.0101061895903
    """
    material = material or MATERIAL

    delta = get_delta(energy, material, density)
    focal_length = radius / 2.0 / delta
    return focal_length


def calc_focal_length(energy, lens_set, material=None, density=None):
    """
    Calculate the Focal Length for certain lenses configuration and energy.

    Parameters
    ----------
    energy : number
        Beam Energy in KeV
    lens_set : list
        [numer1, lensthick1, number2, lensthick2 ...]
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3

    Returns
    -------
    focal_length : float

    Examples
    --------
    >>> calc_focal_length(8, [1, 0.3, 2, 0.4], 'Be')
    11256.040424758363
    """
    material = material or MATERIAL

    f_tot_inverse = 0
    if isinstance(lens_set, int):
        try:
            lens_set = get_lens_set(lens_set)
        except Exception as ex:
            logger.error('When calling get_lens_set error occurred: %s', ex)
            raise ex
    lens_set = (list(zip(lens_set[::2], lens_set[1::2])))
    for num, radius in lens_set:
        if radius is not None:
            fln = calc_focal_length_for_single_lens(energy, radius,
                                                    material, density)
            f_tot_inverse += num/fln

    return 1.0 / f_tot_inverse


def calc_beam_fwhm(energy, lens_set, distance=None, source_distance=None,
                   material=None, density=None,
                   fwhm_unfocused=None, printsummary=True):
    """
    Calculate beam Full Width at the Half Maximum.

    Calculate beam parameters for certain lenses configuration and energy at a
    given distance. FWHM -  Full Width at the Half Maximum Optionally some
    other parameters can be set.

    Parameters
    ----------
    energy : number
        Beam Energy in KeV
    lens_set : list
        [numer1, lensthick1, number2, lensthick2...]
    distance : float
        Distance from the lenses to the sample is 3.852 m at XPP. (in meters)
    source_distance : float, optional
        Distance from source to lenses. This is about 160 m at XPP. (in meters)
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3
    fwhm_unfocused : float, optional
        This is about 400 microns at XPP.
        This is about 900 microns at MEC.
        Radial size of x-ray beam before focusing in meters.
    printsummary : bool, optional
        Prints summary of parameters/calculations if `True`.

    Returns
    -------
    size_fwhm : float
        Beam Full Width at the Half Maximum in meters.

    Examples
    --------
    >>> calc_beam_fwhm(energy=9, lens_set=[2, 0.03, 4, 0.002], distance=4,
                       material='Be', fwhm_unfocused=800e-6)
    0.0008373516816325981
    """
    fwhm_unfocused = fwhm_unfocused or FWHM_UNFOCUSED
    distance = distance or DISTANCE
    material = material or MATERIAL
    # Focal length for certain lenses configuration and energy.
    focal_length = calc_focal_length(energy, lens_set, material, density)

    # Use lens makers equation to find distance to image of source.
    if source_distance is not None:
        focal_length = 1 / (1 / focal_length - 1 / source_distance)

    lam = photon_to_wavelength(energy) * 1e-9

    # The w parameter used in the usual formula is 2 * sigma.
    w_unfocused = gaussian_fwhm_to_sigma(fwhm_unfocused) * 2

    # Assuming gaussian beam divergence = w_unfocused/f we can obtain.
    waist = lam / np.pi * focal_length / w_unfocused
    rayleigh_range = np.pi * waist ** 2 / lam
    size = waist * np.sqrt(1.0 + (distance - focal_length) ** 2.0
                           / rayleigh_range ** 2)

    size_fwhm = gaussian_sigma_to_fwhm(size) / 2.0

    if printsummary:
        logger.info("FWHM at lens   : %.3e", fwhm_unfocused)
        logger.info("waist          : %.3e", waist)
        logger.info("waist FWHM     : %.3e",
                    waist * FWHM_SIGMA_CONVERSION / 2.0)
        logger.info("rayleigh_range : %.3e", rayleigh_range)
        logger.info("focal length   : %.3e", focal_length)
        logger.info("size           : %.3e", size)
        logger.info("size FWHM      : %.3e", size_fwhm)

    return size_fwhm


def calc_distance_for_size(size_fwhm, lens_set, energy,
                           fwhm_unfocused=None):
    """
    Calculate the distance for size.

    Parameters
    ----------
    size_fwhm : float
        Beam Full Width at the Half Maximum in meters.
    lens_set : list
        [numer1, lensthick1, number2, lensthick2...]
    energy : number
        Beam Energy in KeV
    fwhm_unfocused : float, optional
        This is about 400 microns at XPP.
        Radial size of x-ray beam before focusing in meters.

    Returns
    -------
    distance : float
        Distance in meters

    Examples
    --------
    >>> calc_distance_for_size(0.023, [2, 0.03, 4, 0.002], 8, 0.078)
    array([32.00383702, 58.77068253])
    """
    fwhm_unfocused = fwhm_unfocused or FWHM_UNFOCUSED

    size = gaussian_fwhm_to_sigma(size_fwhm) * 2.0
    focal_length = calc_focal_length(energy, lens_set, 'Be', density=None)

    lam = photon_to_wavelength(energy) * 1e-9

    # The w parameter used in the usual formula is 2 * sigma.
    w_unfocused = gaussian_fwhm_to_sigma(fwhm_unfocused) * 2.0

    # Assuming gaussian beam divergence = w_unfocused/f we can obtain.
    waist = lam / np.pi * focal_length / w_unfocused
    rayleigh_range = np.pi * waist ** 2 / lam

    distance = (np.sqrt((size / waist) ** 2 - 1) * np.asarray([-1.0, 1.0])
                * rayleigh_range) + focal_length

    return distance


def calc_lens_aperture_radius(radius, disk_thickness=None,
                              apex_distance=None):
    """
    Calculate the lens aperture radius.

    It is of importance to optimize which lens radius to use at a specific
    photon energy.

    Parameters
    ----------
    radius : float
    disk_thickness : float, optional
        Disk Thickness in meters. Defaults to 1.0e-3.
    apex_distance : float, optional
        Apex Distance in meters. Defaults to 30e-6

    Returns
    -------
    aperture_radius : float

    Examples
    --------
    >>> calc_lens_aperture_radius(radius=4.0, disk_thickness=1e-3,
                                  apex_distance=30e-6)
    0.06228964600958975
    """
    disk_thickness = disk_thickness or DISK_THICKNESS
    apex_distance = apex_distance or APEX_DISTANCE

    aperture_radius = np.sqrt(radius * (disk_thickness - apex_distance))
    return aperture_radius


def calc_trans_for_single_lens(energy, radius, material=None, density=None,
                               fwhm_unfocused=None,
                               disk_thickness=None,
                               apex_distance=None):
    """
    Calculate the transmission for a single lens.

    Parameters
    ----------
    energy : number
        Beam Energy in KeV
    radius : float
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3
    fwhm_unfocused : float, optional
        This is about 400 microns at XPP.
        Radial size of x-ray beam before focusing in meters.
    disk_thickness : float, optional
        Disk Thickness in meters. Defaults to 1.0e-3.
    apex_distance : float, optional
        Apex Distance in meters. Defaults to 30e-6.

    Returns
    -------
    transmission : float
        Transmission for a single lens

    Examples
    --------
    >>> calc_trans_for_single_lens(energy=8, radius=0.03, material='Be',
                                   density=None, fwhm_unfocused=800e-6,
                                   disk_thickness=1.0e-3, apex_distance=30e-6)
    0.9921954096643786
    """
    material = material or MATERIAL
    fwhm_unfocused = fwhm_unfocused or FWHM_UNFOCUSED
    disk_thickness = disk_thickness or DISK_THICKNESS
    apex_distance = apex_distance or APEX_DISTANCE

    # mu = mass attenuation coefficient?
    mu = 1.0 / get_att_len(energy, material=material, density=None)

    sigma = gaussian_fwhm_to_sigma(fwhm_unfocused)
    # TODO: what is S - Responsivity of a lens?
    S = (sigma ** (-2.0) + 2.0 * mu / radius) ** (-0.5)
    aperture_radius = calc_lens_aperture_radius(radius=radius,
                                                disk_thickness=disk_thickness,
                                                apex_distance=apex_distance)

    transmission = ((S ** 2 / sigma ** 2) * np.exp(-mu * apex_distance)
                    * (1 - np.exp(-(aperture_radius ** 2.0) / (2.0 * S ** 2))))
    return transmission


def calc_trans_lens_set(energy, lens_set, material=None, density=None,
                        fwhm_unfocused=None,
                        disk_thickness=None,
                        apex_distance=None):
    """
    Calculate  the transmission of a lens set.

    These would allow us to estimate the total transmission of the lenses.

    Parameters
    ----------
    energy : number
        Beam Energy in KeV
    lens_set : list
        [numer1, lensthick1, number2, lensthick2...]
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3
    fwhm_unfocused : float, optional
        This is about 400 microns at XPP. Default = 900e-6
        Radial size of x-ray beam before focusing in meters.
    disk_thickness : float, optional
        Disk Thickness in meters. Defaults to 1.0e-3.
    apex_distance : float, optional
        Apex Distance in meters. Defaults to 30e-6.

    Returns
    -------
    transmission : float
        Transmission for a set of lens

    Examples
    --------
    >>> calc_trans_lens_set(energy=8, lens_set=[1, 0.03, 4, 0.02],
                            material='Be', density=None,
                            fwhm_unfocused=400e-6)
    0.955752311215339
    """
    material = material or MATERIAL
    fwhm_unfocused = fwhm_unfocused or FWHM_UNFOCUSED
    disk_thickness = disk_thickness or DISK_THICKNESS
    apex_distance = apex_distance or APEX_DISTANCE

    apex_distance_tot = 0
    radius_total_inv = 0
    if isinstance(lens_set, int):
        try:
            lens_set = get_lens_set(lens_set)
        except Exception as ex:
            logger.error('When calling get_lens_set error occurred: %s', ex)
            raise ex
    lens_set = (list(zip(lens_set[::2], lens_set[1::2])))

    radius_total_inv = sum(num / radius for num, radius in lens_set)
    apex_distance_tot = sum(num * apex_distance for num, _ in lens_set)
    radius_aperture = min(np.sqrt(radius * (disk_thickness - apex_distance))
                          for _, radius in lens_set)

    radius_total = 1.0 / radius_total_inv
    equivalent_disk_thickness = (radius_aperture ** 2 / radius_total
                                 + apex_distance_tot)

    transmission_total = calc_trans_for_single_lens(energy, radius_total,
                                                    material, density,
                                                    fwhm_unfocused,
                                                    equivalent_disk_thickness,
                                                    apex_distance_tot)
    return transmission_total


def calc_lens_set(energy, size_fwhm, distance, n_max=25, max_each=5,
                  lens_radii=None,
                  fwhm_unfocused=None, eff_rad0=None):
    """
    Calculate lens set.

    Parameters
    ----------
    energy : number
        Beam Energy in KeV
    size_fwhm : float
        Beam Full Width at the Half Maximum. (in meters)
    distance : float
        Distance from the lenses to the sample. (in meters)
    n_max : int, optional
        Number of lenses.
    max_each : int, optional
    lens_radii : list, optional
          [numer1, lensthick1, number2, lensthick2...]
    fwhm_unfocused : float, optional
        This is about 400 microns at XPP. Defaults to 500e-6.
        Radial size of x-ray beam before focusing in meters.
    eff_rad0 : float, optional

    Returns
    -------
    lens_sets : tuple
        Lens sets

    Examples
    --------
    >>> calc_lens_set(energy=7, size_fwhm=0.54, distance=3)
    """
    lens_radii = lens_radii or LENS_RADII
    fwhm_unfocused = fwhm_unfocused or FWHM_UNFOCUSED

    nums = product(*([list(range(max_each + 1))] * len(lens_radii)))
    sets = []
    sizes = []
    eff_rads = []
    foc_lens = []
    for num in nums:
        lens_set = []
        if sum(num) <= n_max and sum(num) > 0:
            if eff_rad0 is None:
                teff_rad_inv = 0
            else:
                teff_rad_inv = 1 / eff_rad0
            for tn, tl in zip(num, lens_radii):
                lens_set += [tn, tl]
                teff_rad_inv += tn / tl
            teff_rad = np.round(1 / teff_rad_inv, 6)
            if teff_rad in eff_rads:
                ind = eff_rads.index(teff_rad)
                if sum(sets[ind]) > sum(num):
                    sets[ind] = num
                else:
                    continue
            else:
                eff_rads.append(teff_rad)
                sets.append(num)
                sizes.append(calc_beam_fwhm(energy, lens_set + [1, eff_rad0],
                                            distance=distance,
                                            source_distance=None,
                                            fwhm_unfocused=fwhm_unfocused,
                                            printsummary=False))
                foc_lens.append(calc_focal_length(energy,
                                                  lens_set + [1, eff_rad0]))

    sizes = np.asarray(sizes)
    sets = np.asarray(sets)
    foc_lens = np.asarray(foc_lens)
    indsort = (np.abs(sizes - size_fwhm)).argsort()

    lens_sets = (sets[indsort, :],
                 np.asarray(eff_rads)[indsort],
                 sizes[indsort],
                 foc_lens[indsort])
    return lens_sets


def find_radius(energy, distance=None, material=None, density=None):
    """
    Find the radius of curvature.

    Find the radius of curvature of the lens that would focus the energy at the
    distance.

    Parameters
    ----------
    energy : number
        Beam Energy
    distance : float, optional
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3

    Returns
    -------
    radius : float
        Radius of curvature in meters.

    Examples
    --------
    >>> find_radius(energy=8, distance=4.0, material='Be', density=None)
    4.2643770090253954e-05
    """
    distance = distance or DISTANCE
    material = material or MATERIAL

    delta = get_delta(energy, material, density)
    radius = distance * 2 * delta
    return radius


def find_energy(lens_set, distance=None, material=None, density=None):
    """
    Find the energy that would focus at a given distance.

    Parameters
    ----------
    lens_set : list
        [numer1, lensthick1, number2, lensthick2...]
    distance : float, optional
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3

    Returns
    -------
    energy : float
        Energy

    Examples
    --------
    >>> find_energy([2, 200e-6, 4, 500e-6], distance=4)
    7.0100555419921875
    """
    distance = distance or DISTANCE
    material = material or MATERIAL

    energy_min = 1.0
    energy_max = 24.0
    energy = (energy_max + energy_min) / 2.0
    abs_diff = 100
    while abs_diff > 0.0001:
        focal_length_min = calc_focal_length(energy_min, lens_set,
                                             material, density)
        focal_length_max = calc_focal_length(energy_max, lens_set,
                                             material, density)
        energy = (energy_max + energy_min) / 2.0
        focal_length = calc_focal_length(energy, lens_set, material, density)
        if (distance < focal_length_max) and (distance > focal_length):
            energy_min = energy
        elif (distance > focal_length_min) and (distance < focal_length):
            energy_max = energy
        else:
            logger.error("somehow failed ...")
            break
        abs_diff = abs(distance - focal_length)
    logger.info("Energy that would focus at a distance of %.3f is %.3f",
                distance, energy)

    return energy


def find_z_pos(energy, lens_set, spot_size_fwhm, material=None,
               density=None, fwhm_unfocused=None):
    """
    Find the Be Lens distances.

    Find the two distances the Be lens needs to be at to get the spotsize in
    the chamber center.

    Parameters
    ----------
    energy : number
        Beam Energy in KeV
    lens_set : list
        [numer1, lensthick1, number2, lensthick2...]
    spot_size_fwhm :
    material : str, optional
        Atomic symbol for element, defaults to 'Be'.
    density : float, optional
        Material density in g/cm^3
    fwhm_unfocused : float, optional
        This is about 400 microns at XPP. Defaults to 500e-6
        Radial size of x-ray beam before focusing in meters.

    Returns
    -------
    z_position : tuple
        (z1, z2)

    Examples
    --------
    >>> lens_set = [2, 200e-6, 4, 500e-6]
    >>> find_z_pos(energy=8, lens_set=lens_set, spot_size_fwhm=0.09,
                   material='Be', density=None, fwhm_unfocused=800e-6)
    (-2339.797291538794, 2350.2195511913483)
    """
    material = material or MATERIAL
    fwhm_unfocused = fwhm_unfocused or FWHM_UNFOCUSED

    focal_length = calc_focal_length(energy, lens_set, material, density)

    lam = photon_to_wavelength(energy) * 1e-9
    # The w parameter used in the usual formula is 2 * sigma.
    w_unfocused = gaussian_fwhm_to_sigma(fwhm_unfocused) * 2
    waist = lam / np.pi * focal_length / w_unfocused
    rayleigh_range = np.pi * waist ** 2 / lam

    logger.info("waist          : %.3e", waist)
    logger.info("waist FWHM     : %.3e", waist * FWHM_SIGMA_CONVERSION / 2.0)
    logger.info("rayleigh_range : %.3e", rayleigh_range)
    logger.info("focal length   : %.3e", focal_length)

    w = gaussian_fwhm_to_sigma(spot_size_fwhm) * 2
    delta_z = rayleigh_range * np.sqrt((w / waist) ** 2 - 1)
    z1 = focal_length - delta_z
    z2 = focal_length + delta_z
    z_position = (z1, z2)
    return z_position


def plan_set(energy, z_offset, z_range, beam_size_unfocused, size_horizontal,
             size_vertical=None, exclude=None, max_tot_number_of_lenses=25,
             max_each=5, focus_before_sample=False):
    """
    Macro to help plan for what lens set to use.

    Parameters
    ----------
    energy : number
        Beam energy in KeV
    z_offset : float
        Distance from sample to lens_z=0 in meters
    z_range : list
        Array or tuple of 2 values: minimum and maximum z pos in meters
    beam_size_unfocused : float
        Radial size of x-ray beam before focusing in meters
    size_horizontal : array-like
    size_vertical : None, int or tuple of ints, optional
    exclude : list
        List with excluded sets of lenses
    max_tot_number_of_lenses : int
        Number of lenses.
    max_each : int
    focus_before_sample : bool

    Examples
    --------
    >>> plan_set(energy=1, z_offset=-10, z_range=[1, 40],
                           beam_size_unfocused=3, size_horizontal=9,
                           size_vertical=None, exclude=[],
                           max_tot_number_of_lenses=1,
                           max_each=5, focus_before_sample=False)

    N   foc_len/m   Min/um   Max/um   Trans/%  Set
    0  0.07 466994229.0 2112064677.4   0.0  1 x 50um
    1  0.14 234997114.5 1057532338.7   0.0  1 x 100um
    2  0.28 118998557.3 530266169.4   0.0  1 x 200um
    3  0.43 80332371.5 354510779.6   0.0  1 x 300um
    4  0.71 49399422.9 213906467.7   0.0  1 x 500um
    5  1.42 26199711.5 108453233.9   0.0  1 x 1000um
    6  2.13 18466474.3 73302155.9   0.0  1 x 1500um
    7  2.84 14599855.7 55726616.9   0.0  1 x 2000um
    8  4.27 10733237.2 38151078.0   0.0  1 x 3000um

    Where:
    `foc_len` - is the focal length in um
    `Min` - Beam full width at half maximum for minimum z position
    `Max` - Beam full width at half maximum for maximum z position
    `Trans` - is the lens transmission in %.

    Notes
    -----
    When providing a large number to `max_tot_number_of_lenses`,
    it will take some time to run the calculations, for example:
    `max_tot_number_of_lenses=25` will take ~ 2 min
    """
    global _plan_set_test_res
    exclude = exclude or []

    if None in (z_offset, z_range, beam_size_unfocused):
        logger.error('Cannot plan_set. At least one of z_offset,'
                     ' z_range, beam_size_unfocused not defined.')
        return
    focus_before_sample = int(focus_before_sample)
    if size_vertical is not None:
        size_calc_1 = np.max(size_horizontal, size_vertical)
    else:
        size_calc_1 = size_horizontal

    t_lens_radii = LENS_RADII
    for rad in exclude:
        t_lens_radii.remove(rad)

    sets, effrads, sizes, foc_lens = calc_lens_set(
        energy=energy,
        size_fwhm=size_calc_1,
        distance=z_offset,
        n_max=max_tot_number_of_lenses,
        max_each=max_each,
        lens_radii=LENS_RADII,
    )

    distances = np.asarray(
        [
            calc_distance_for_size(
                size_calc_1,
                list(chain(*list(zip(set, t_lens_radii)))),
                energy=energy,
                fwhm_unfocused=beam_size_unfocused,
            )[focus_before_sample]
            for set in sets
        ]
    )

    good_sets = np.logical_and(
        distances > np.min(z_offset + np.asarray(z_range)),
        distances < np.max(z_offset + np.asarray(z_range))
    )

    sets = sets[good_sets]
    size_range_min = np.asarray(
        [
            calc_beam_fwhm(
                energy,
                list(chain(*list(zip(set, t_lens_radii)))),
                distance=z_offset - min(z_range),
                fwhm_unfocused=beam_size_unfocused,
                printsummary=False,
            )
            for set in sets
        ]
    )

    size_range_max = np.asarray([
        calc_beam_fwhm(
            energy,
            list(chain(*list(zip(set, t_lens_radii)))),
            distance=z_offset - max(z_range),
            fwhm_unfocused=beam_size_unfocused,
            printsummary=False,
        ) for set in sets])

    sizes = sizes[good_sets]
    effrads = effrads[good_sets]
    distances = distances[good_sets]
    foclens = foc_lens[good_sets]
    transms = np.asarray(
        [lens_transmission(radius=ter, fwhm=beam_size_unfocused, energy=energy)
            for ter in effrads])
    Nlenses_s = np.sum(sets, 1)

    # used for testing
    num = []
    f_m = []
    min_um = []
    max_um = []
    t_percent = []

    resstring = " N   foc_len/m   Min/um   Max/um   Trans/%  Set\n"
    zips = list(
        zip(sets, size_range_min, size_range_max,
            effrads, transms, Nlenses_s, foclens)
    )

    t = "{:2d} {:5.2f} {:8.1f} {:8.1f} {:5.1f}  "
    for n, z in enumerate(zips):
        the_set, sizemin, sizemax, eff_rad, transm, n_lenses, foclen = z
        logger.debug('eff_rad: %s', eff_rad)
        logger.debug('n_lenses: %s', n_lenses)

        num.append(n)
        f_m.append(foclen)
        min_um.append(sizemin * 1e6)
        max_um.append(sizemax * 1e6)
        t_percent.append(transm * 100)

        resstring += t.format(n, foclen, sizemin * 1e6,
                              sizemax * 1e6, transm * 100)
        resstring += ", ".join(
            [
                "%d x %dum" % (setLensno, t_lens_radii[m] * 1e6)
                for m, setLensno in enumerate(the_set)
                if setLensno > 0
            ]
        )
        resstring += "\n"
    logger.info('\n %s', resstring)
    _plan_set_test_res = num, f_m, min_um, max_um, t_percent


def lens_transmission(radius, fwhm, energy, num=1, id_material="IF1",
                      lens_thicknes=None):
    """
    Find the CRL (Compound Refractive Lens) transmission.

    Parameters
    ----------
    radius : float
        Effective radius of curvature in meters
    fwhm : float
        Incident beam size on the lens in meters
    num : int
        Number of lenses in the stack
    energy : number
        Photon energy in KeV
    id_material : str
        Lens material, Defaults to `IF1`
    lens_thickness : float
        Lens thickness in meters

    Returns
    -------
    trans : float
        Lens Transmission
    """
    lens_thicknes = lens_thicknes or APEX_DISTANCE
    id_material = chemical_name_to_formula.get(id_material, id_material)
    waist = 2 * gaussian_fwhm_to_sigma(fwhm)
    x = np.linspace(-2 * fwhm, 2 * fwhm, 101)
    y = x
    intensity = np.zeros((len(x), len(y)))
    thickness = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            intensity[i, j] = (
                np.abs(np.exp(-2 * (x[i] ** 2 + y[j] ** 2) / waist ** 2))
                * 2
                / waist ** 2
                / np.pi
                * (x[2] - x[1]) ** 2
            )
            thickness[i, j] = ((x[i] ** 2 + y[j] ** 2) / radius + num *
                               lens_thicknes)
    d = density.get(id_material)
    att_length = get_att_len(energy, id_material, d)
    trans_intensity = intensity * np.exp(-thickness / att_length)
    trans = np.sum(trans_intensity)
    return trans



"""
Calculation utilities related to the LCLS PMPS system.

Here you will find functions related to eV bitmasks.
"""
import math
from typing import Optional

# Source for these values:
# lcls-twincat-pmps/PMPS/GVLs/PMPS_GVL.TcGVL
LFE = [
    1.0E3,
    1.7E3,
    2.1E3,
    2.5E3,
    3.8E3,
    4.0E3,
    5.0E3,
    7.0E3,
    7.5E3,
    7.7E3,
    8.9E3,
    10.0E3,
    11.1E3,
    12.0E3,
    13.0E3,
    13.5E3,
    14.0E3,
    16.9E3,
    18.0E3,
    20.0E3,
    22.0E3,
    24.0E3,
    25.0E3,
    25.5E3,
    26.0E3,
    27.0E3,
    28.0E3,
    28.5E3,
    29.0E3,
    30.0E3,
    60.0E3,
    90.0E3,
]

KFE = [
    100,
    250,
    270,
    350,
    400,
    450,
    480,
    530,
    680,
    730,
    850,
    1.10E3,
    1.15E3,
    1.25E3,
    1.45E3,
    1.50E3,
    1.55E3,
    1.65E3,
    1.70E3,
    1.75E3,
    1.82E3,
    1.85E3,
    2.00E3,
    2.20E3,
    2.50E3,
    2.80E3,
    3.00E3,
    3.15E3,
    3.50E3,
    4.00E3,
    5.30E3,
    7.00E3
]


def select_bitmask_boundaries(line: str) -> list[float]:
    """
    Given a line, select the bitmask boundaries to use.

    These are hard-coded for now but in theory this could be loaded from
    a database, an EPICS PV, or some other source.

    Parameters
    ----------
    line: str
        String representation of which line's bitmask to use.
        If the string begins with "l" or "h" (lfe, hxr), we'll
        use the hard-xray bitmask.
        If the string begins with "k" or "s" (kfe, sxr), we'll
        use the soft-xray bitmask.

    Returns
    -------
    boundaries: list of floats
    """
    if line.lower()[0] in ['l', 'h']:
        return LFE
    if line.lower()[0] in ['k', 's']:
        return KFE
    raise ValueError(f'{line} is neither lfe or kfe!.')


def get_bitmask(
    lower: float,
    upper: float,
    allow: bool,
    line: str,
    bounds: Optional[list[float]] = None,
) -> int:
    """
    Given a range of eV values, calculate the appropriate pmps bitmask.

    This saves you the effort of checking up on the eV ranges and
    remembering how the bitmask is assembled.

    If the lower or upper fall within a range (not on the border), then that
    range is always considered unsafe.

    The rules for the bitmasks are:
        - The nth bit of the bitmask represents a range from the n-1th value
          to the nth value
        - If the bit is 1, the range is allowed and is safe.
        - If the bit is 0, the range is not allowed and is not safe.
        - eVs above the highest boundary are never safe
        - negative eVs are never safe

    Some examples:
        - bitmask 00 -> no beam range is safe
        - bitmask 01 -> only the lowest boundary and below are safe
        - bitmask 11 -> only the two lowest boundaries and below are safe
        - bitmask 10 -> only the space between the lowest and second-lowest
                        boundaries are safe.
        - bitmask all 1s -> every eV is OK except for above the highest
                            boundary and except for negative eV

    Parameters
    ----------
    lower: number
        The value in eV for the lower bound of the range.

    upper: number
        The value in eV for the upper bound of the range.

    allow: bool
        True if we want a bitmask that only includes this range,
        False if we want a bitmask that only excludes this range.

    line: str
        String representation of which line's bitmask to use.
        If the string begins with "l" or "h" (lfe, hxr), we'll
        use the hard-xray bitmask.
        If the string begins with "k" or "s" (kfe, sxr), we'll
        use the soft-xray bitmask.

    bounds: list of numbers, optional
        Custom boundaries to use instead of the default soft-xray
        or hard-xray lines. Useful for testing.

    Returns
    -------
    bitmask: int
    """
    # Help the user if they accidently pass a negative
    if lower < 0 or upper < 0:
        raise ValueError('get_bitmask is only valid for positive inputs')

    # Be lenient on input args
    if upper < lower:
        lower, upper = upper, lower

    if allow:
        bounds = bounds or select_bitmask_boundaries(line)
        bitmask = 0

        prev = 0
        for bit, ev in enumerate(bounds):
            if lower <= prev and upper >= ev:
                bitmask += 2**bit
            prev = ev
        return bitmask

    # An exclusion range is just two inclusion ranges
    else:
        lower_range = get_bitmask(0, lower, True, line, bounds=bounds)
        upper_range = get_bitmask(upper, math.inf, True, line, bounds=bounds)
        return lower_range | upper_range


def check_bitmask(
    energy: float,
    bitmask: int,
    line: str,
    bounds: Optional[list[float]] = None,
) -> bool:
    """
    Given an energy and a bitmask, tell us if our energy is allowed.

    This is the same calculation the PMPS is doing internally to determine
    if it is safe for beam to proceed.

    Parameters
    ----------
    energy: number
        The value in eV for the energy to check.

    bitmask: int
        The bits to check against. Typically an output of `get_bitmask`.

    line: str
        String representation of which line's bitmask to use.
        If the string begins with "l" or "h" (lfe, hxr), we'll
        use the hard-xray bitmask.
        If the string begins with "k" or "s" (kfe, sxr), we'll
        use the soft-xray bitmask.

    bounds: list of numbers, optional
        Custom boundaries to use instead of the default soft-xray
        or hard-xray lines. Useful for testing.

    Returns
    -------
    energy_allowed: bool
        True if the energy is allowed.
    """
    bounds = bounds or select_bitmask_boundaries(line)

    # Boundary energy exists in two ranges, so we can get two answers
    answers = []
    prev = 0
    for bit, ev in enumerate(bounds):
        if prev <= energy <= ev:
            ok = bool((bitmask >> bit) % 2)
            answers.append(ok)
        prev = ev

    if not answers:
        # We get here if energy is negative or too large
        return False
    else:
        return all(answers)


def check_actual_range(
    lower: float,
    upper: float,
    allow: bool,
    line: str,
    bounds: Optional[list[float]] = None,
) -> tuple[float, float]:
    """
    Returns the actual effective range given bitmask precision.

    Because of the granularity of the bitmask, most range specifications
    exclude more energy values than requested.

    Parameters
    ----------
    lower: number
        The value in eV for the lower bound of the range.

    upper: number
        The value in eV for the upper bound of the range.

    allow: bool
        True if we want a bitmask that only includes this range,
        False if we want a bitmask that only excludes this range.

    line: str
        String representation of which line's bitmask to use.
        If the string begins with "l" or "h" (lfe, hxr), we'll
        use the hard-xray bitmask.
        If the string begins with "k" or "s" (kfe, sxr), we'll
        use the soft-xray bitmask.

    bounds: list of numbers, optional
        Custom boundaries to use instead of the default soft-xray
        or hard-xray lines. Useful for testing.

    Returns
    -------
    ranges: tuple
        A (lower, upper) pair that represents a range of allowed
        (or forbidden) energy values. The endpoints of the range
        are considered unsafe.
    """
    bitmask = get_bitmask(lower, upper, allow, line, bounds=bounds)
    if not allow:
        bitmask = ~bitmask
    lowest = math.inf
    highest = -math.inf
    updated_range = False

    prev = 0
    for bit, ev in enumerate(bounds):
        ok = bool((bitmask >> bit) % 2)
        if ok:
            lowest = min(lowest, prev)
            highest = max(highest, ev)
            updated_range = True
        prev = ev
    if updated_range:
        return (lowest, highest)
    else:
        # The range is empty: return an empty range instead of inf inf.
        return (lower, lower)


def describe_bitmask(
    bitmask: int,
    line: str,
    bounds: Optional[list[float]] = None,
) -> None:
    """
    Print a text description of a bitmask.

    This will describe what the bitmask means.

    Parameters
    ----------
    bitmask : int
        The bits to describe. Typically an output of `get_bitmask`.

    line : str
        String representation of which line's bitmask to use.
        If the string begins with "l" or "h" (lfe, hxr), we'll
        use the hard-xray bitmask.
        If the string begins with "k" or "s" (kfe, sxr), we'll
        use the soft-xray bitmask.

    bounds : list of numbers, optional
        Custom boundaries to use instead of the default soft-xray
        or hard-xray lines. Useful for testing.
    """
    lines = get_bitmask_desc(bitmask=bitmask, line=line, bounds=bounds)
    print('\n'.join(lines))


def get_bitmask_desc(
    bitmask: int,
    line: str,
    bounds: Optional[list[float]] = None,
) -> list[str]:
    """
    Return a text description of a bitmask.

    This will describe what the bitmask means.

    Parameters
    ----------
    bitmask : int
        The bits to describe. Typically an output of `get_bitmask`.

    line : str
        String representation of which line's bitmask to use.
        If the string begins with "l" or "h" (lfe, hxr), we'll
        use the hard-xray bitmask.
        If the string begins with "k" or "s" (kfe, sxr), we'll
        use the soft-xray bitmask.

    bounds : list of numbers, optional
        Custom boundaries to use instead of the default soft-xray
        or hard-xray lines. Useful for testing.

    Returns
    -------
    desc : list of str
        Description lines that describe what the bitmask means.
    """
    bounds = bounds or select_bitmask_boundaries(line)
    lines = []

    width = 0
    for bound in bounds:
        width = max(width, len(str(bound)))

    prev = 0
    for bit, ev in enumerate(bounds):
        val = (bitmask >> bit) % 2
        count = bit + 1
        if val:
            text = 'allowed'
        else:
            text = 'disallowed'
        line = f'Bit {count:2}: {val} ({prev:{width}}, {ev:{width}}) {text}'
        lines.append(line)
        prev = ev
    return lines



import logging

import numpy as np
import pytest

from pcdscalc import common

logger = logging.getLogger(__name__)


@pytest.mark.parametrize('angle, expected', [
                         pytest.param(10, 0.984807753012208),
                         pytest.param(180, -1.0),
                         pytest.param(360, 1.0),
                         pytest.param(45, 0.7071067811865476)
                         ])
def test_cosd(angle, expected):
    res = common.cosd(angle)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


@pytest.mark.parametrize('angle, expected', [
                         pytest.param(10, 0.17364817766693033),
                         pytest.param(180, 1.2246467991473532e-16),
                         pytest.param(360, -2.4492935982947064e-16),
                         pytest.param(45, 0.7071067811865476),
                         pytest.param(270, -1.0)
                         ])
def test_sind(angle, expected):
    res = common.sind(angle)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


@pytest.mark.parametrize('angle, expected', [
                         pytest.param(45, 1),
                         pytest.param(60, np.sqrt(3)),
                         pytest.param(30, 1/np.sqrt(3))
                         ])
def test_tand(angle, expected):
    res = common.tand(angle)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


@pytest.mark.parametrize('x, expected', [
                         pytest.param(-1, -90),
                         pytest.param(1, 90),
                         pytest.param(0.2, 11.536959032815489)
                         ])
def test_asind(x, expected):
    res = common.asind(x)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


@pytest.mark.parametrize('x, expected', [
                         pytest.param(-1, 180),
                         pytest.param(1, 0),
                         pytest.param(0, 90)
                         ])
def test_acosd(x, expected):
    res = common.acosd(x)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


@pytest.mark.parametrize('x, expected', [
                         pytest.param(9, 83.66),
                         pytest.param(1, 45),
                         pytest.param(0, 0)
                         ])
def test_atand(x, expected):
    res = common.atand(x)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


@pytest.mark.parametrize('energy, expected', [
                         pytest.param(2e3, 6.1992e-10),
                         pytest.param(10e3, 1.23984e-10),
                         pytest.param(9e3, 1.3776e-10),
                         pytest.param(8e3, 1.5498e-10)
                         ])
def test_energy_to_wavelength(energy, expected):
    # tested agains the old code
    res = common.energy_to_wavelength(energy)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


@pytest.mark.parametrize('wavelength, expected', [
                         pytest.param(100, 1.23984e-08),
                         pytest.param(9, 1.3775999999999998e-07),
                         pytest.param(7, 1.7712e-07),
                         pytest.param(0.8, 1.5498e-06)
                         ])
def test_wavelength_to_energy(wavelength, expected):
    # tested agains the old code
    res = common.wavelength_to_energy(wavelength)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)



import logging

import numpy as np
import pytest

from pcdscalc import diffraction

logger = logging.getLogger(__name__)


@pytest.mark.parametrize('energy, material, reflection, expected', [
                         pytest.param(10e3, 'Si', (1, 1, 1),
                                      (11.402710639982848, 713.4828146545175)),
                         pytest.param(10e3, 'C', (1, 1, 1),
                                      (17.51878596767417, 427.8469911590626)),
                         pytest.param(9e3, 'Si', (2, 2, 2),
                                      (26.061879662584946, 233.3438968335023)),
                         pytest.param(9e3, 'C', (2, 2, 2),
                                      (41.98453277509927, 31.69504158241886))
                         ])
def test_get_geometry(energy, material, reflection, expected):
    # tested agains the old code
    res = diffraction.get_lom_geometry(energy, material, reflection)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res).all()


@pytest.mark.parametrize('material_id, hkl, energy, expected', [
                         pytest.param('Si', (1, 1, 1), 10e3, 11.4027106399828),
                         pytest.param('Si', (1, 1, 1), 7e3, 16.40552015690347),
                         pytest.param('C', (1, 1, 1), 10e3, 17.51878596767417),
                         pytest.param('C', (2, 2, 2), 10e3, 37.01592462373143),
                         pytest.param('Si', (2, 2, 2), 7e3, 34.39310890947986),
                         pytest.param('C', (2, 2, 2), 9.8e3, 37.90277660759642)
                         ])
def test_bragg_angle(material_id, hkl, energy, expected):
    # tested agains the old code
    res = diffraction.bragg_angle(material_id, hkl, energy)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


@pytest.mark.parametrize('material_id, hkl, expected', [
                         pytest.param('Si', (1, 1, 1), 3.1356011476493755e-10),
                         pytest.param('Si', (2, 2, 2), 1.5678005738246877e-10),
                         pytest.param('C', (1, 1, 1), 2.059408410199395e-10),
                         pytest.param('C', (2, 2, 2), 1.0297042050996975e-10)
                         ])
def test_d_space(material_id, hkl, expected):
    # tested agains the old code
    res = diffraction.d_space(material_id, hkl)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)



import logging
import os
from unittest.mock import patch

import numpy as np
import pytest

from pcdscalc import be_lens_calcs

logger = logging.getLogger(__name__)

PATH = os.path.dirname(__file__) + '/test_lens_sets/lens_set'
BAD_PATH = 'bad/path/to/lens_set'

SETS_SAMPLE = [[3, 0.0001, 1, 0.0002],
               [1, 0.0001, 1, 0.0003, 1, 0.0005],
               [2, 0.0001, 1, 0.0005]]


def test_configure_defaults():
    # make sure is not the value we want to test
    be_lens_calcs.FWHM_UNFOCUSED = 500e-6
    be_lens_calcs.configure_defaults(fwhm_unfocused=800e-6)
    assert be_lens_calcs.FWHM_UNFOCUSED == 800e-6


def test_configure_lens_set_bad_path():
    with pytest.raises(FileNotFoundError):
        be_lens_calcs.configure_lens_set_file(BAD_PATH)


def test_get_lens_set_with_bad_path():
    with pytest.raises(FileNotFoundError):
        be_lens_calcs.get_lens_set(1, BAD_PATH)


def test_get_lens_set_with_empty_file():
    be_lens_calcs.set_lens_set_to_file("", PATH, False)
    # file should be empty
    with pytest.raises(ValueError):
        be_lens_calcs.get_lens_set(1, PATH)


def test_get_lens_set_file_one_set():
    first_set = [3, 0.0001, 1, 0.0002]
    be_lens_calcs.set_lens_set_to_file(first_set, PATH, False)
    lens_set = be_lens_calcs.get_lens_set(1, PATH)
    logger.debug(f'result: {lens_set}')
    assert lens_set == first_set


def test_lens_file_with_multiple_sets():
    be_lens_calcs.set_lens_set_to_file(SETS_SAMPLE, PATH, False)
    lens_set = be_lens_calcs.get_lens_set(2, PATH)
    expected = [1, 0.0001, 1, 0.0003, 1, 0.0005]
    assert expected == lens_set


def test_configure_lens_set_file():
    res = be_lens_calcs.configure_lens_set_file(PATH)
    assert res == os.path.abspath(PATH)
    assert be_lens_calcs.LENS_SET_FILE == res


def test_get_lens_set():
    first_set = [3, 0.0001, 1, 0.0002]
    lens_set = be_lens_calcs.get_lens_set(1, PATH)
    logger.debug(f'result: {lens_set}')
    assert lens_set == first_set


def test_get_lens_set_no_file():
    be_lens_calcs.LENS_SET_FILE = None
    with pytest.raises(ValueError):
        be_lens_calcs.get_lens_set(1)


@pytest.mark.parametrize('energy_sample, expected', [
    pytest.param(8, 0.004810113254120656),
    pytest.param(9, 0.006507409652366966),
    pytest.param(1, 8.946976460379717e-06),
    pytest.param(30, 0.030160746908555823),
    pytest.param(10, 0.008360110829541422),
])
def test_get_att_length(energy_sample, expected):
    # old values (with old get_att_lenght)
    # 0.005949593420831057
    # 0.00874765768614978
    # 9.030173195532919e-06
    # 0.4711761144434822
    # 0.01235515814053393
    att_len = be_lens_calcs.get_att_len(energy_sample, material='Be')
    logger.debug('At energy: %s, Expected: %s,  Received %s',
                 energy_sample, expected, att_len)
    assert np.isclose(expected, att_len)


def test_get_att_len_with_bad_material():
    # should raise an exception, provided unknown element string
    with pytest.raises(ValueError):
        be_lens_calcs.get_att_len(8, 'BEES', 32)


@pytest.mark.parametrize('energy_sample, expected', [
    pytest.param(8, 5.330471261281744e-06),
    pytest.param(9, 4.21084077002856e-06),
    pytest.param(1, 0.0003515107795702183),
    pytest.param(0.1, 0.005621922785393444),
    pytest.param(20, 8.521078856544226e-07),
])
def test_get_delta(energy_sample, expected):
    # old values (using old get_data):
    # 5.326454632470501e-06
    # 4.20757010499706e-06
    # 0.0003524314661392802
    # 0.003807346693343705
    # 8.513802953746819e-07
    delta = be_lens_calcs.get_delta(energy_sample, material='Be')
    logger.debug('At energy: %s, Expected: %s, Received %s',
                 energy_sample, expected, delta)
    assert np.isclose(expected, delta)


def test_get_delta_with_0_energy():
    # should give an exception since can't devide by 0
    with pytest.raises(ZeroDivisionError):
        be_lens_calcs.get_delta(0)


@pytest.mark.parametrize('energy_sample, radius, expected', [
    pytest.param(8, 0.003, 281.40101061895905),
    pytest.param(8, 0.0001, 9.380033687298635),
    pytest.param(9, 0.003, 356.2233962102125),
    pytest.param(1, 0.0001, 0.14224314844948285),
])
def test_calc_focal_length_for_single_lens(energy_sample, radius, expected):
    # old values (with old get_delta):
    # 281.61321244639504
    # 9.387107081546501
    # 356.5002988823755
    # 0.1418715546251483
    focal_length = be_lens_calcs.calc_focal_length_for_single_lens(
        energy_sample, radius)
    logger.debug('Expected: %s, Received: %s', expected, focal_length)
    assert np.isclose(focal_length, expected)


@pytest.mark.parametrize('energy_sample, lens_set, expected', [
    pytest.param(8, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02], 69.39814397219442),
    pytest.param(8, [2, 200e-6, 4, 500e-6], 5.21112982627702),
    pytest.param(9, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02], 87.85058192251857),
    pytest.param(1, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02], 1.0523853990548204),
])
def test_calc_focal_length(energy_sample, lens_set, expected):
    # old values (with old get_delta):
    # 69.45047645294554
    # 5.2150594897480556
    # 87.91887070181892
    # 1.0496361635424505
    fl = be_lens_calcs.calc_focal_length(energy_sample, lens_set)
    logger.debug('Expected: %s, Received: %s', expected, fl)
    assert np.isclose(fl, expected)


def test_calc_focal_length_with_file_lens_set():
    # expected_used_set = [3, 0.0001, 1, 0.0002]
    with patch('pcdscalc.be_lens_calcs.get_lens_set',
               return_value=[3, 0.0001, 1, 0.0002]):
        expected = be_lens_calcs.calc_focal_length(8, [3, 0.0001, 1, 0.0002])
        received = be_lens_calcs.calc_focal_length(8, 1)
        logger.debug('Expected: %s, Received: %s', expected, received)
        assert received == expected


def test_calc_focal_length_with_no_file():
    # make sure the LENS_SET_FILE is none
    # should get an Exception
    be_lens_calcs.LENS_SET_FILE = None
    with pytest.raises(ValueError):
        be_lens_calcs.calc_focal_length(8, 1)


@pytest.mark.parametrize('energy_sample, lens_set, dist, fwhm_unf, expect', [
    pytest.param(8, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02],
                 4, 800e-6, 0.0007539125970041841),
    pytest.param(8, [2, 200e-6, 4, 500e-6],
                 4, 800e-6, 0.00018593024432864825),
    pytest.param(8, [2, 200e-6, 4, 500e-6],
                 3, 500e-6, 0.00021215574911270682),
    pytest.param(8, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02],
                 4, 500e-6, 0.0004712763789963477),
    pytest.param(1, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02],
                 3, 500e-6, 0.0009253340604551429),
])
def test_calc_beam_fwhm(energy_sample, lens_set, dist, fwhm_unf, expect):
    # old values (with old get_delta):
    # 0.000753947185823854
    # 0.00018639295509465447
    # 0.00021237263787327295
    # 0.0004712974533787206
    # 0.0009290673198171924
    fwhm = be_lens_calcs.calc_beam_fwhm(energy=energy_sample,
                                        lens_set=lens_set,
                                        distance=dist,
                                        fwhm_unfocused=fwhm_unf)
    logger.debug('Expected: %s, Received: %s', expect, fwhm)
    assert np.isclose(fwhm, expect)


@pytest.mark.parametrize('energy_sample, lens_set, dist ,'
                         'fwhm_unf, source_dist, expected', [
                             pytest.param(8, [2, 200e-6, 4, 500e-6],
                                          4, 500e-6, 10,
                                          0.0003162095431898309),
                         ])
def test_calc_beam_fwhm_with_source_distance(energy_sample, lens_set, dist,
                                             fwhm_unf, source_dist, expected):
    # old values (with old get_delta):
    # 0.000316498748238284
    fwhm = be_lens_calcs.calc_beam_fwhm(energy=energy_sample,
                                        lens_set=lens_set, distance=dist,
                                        fwhm_unfocused=fwhm_unf,
                                        source_distance=source_dist)
    logger.debug('Expected: %s, Received: %s', expected, fwhm)
    assert np.isclose(fwhm, expected)


@pytest.mark.parametrize('energy, lens_set, fwhm_unf, size_fwhm, expect', [
    pytest.param(8, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02],
                 800e-6, 0.0007539124064936633, [4.00001653, 134.79627142]),
    pytest.param(8, [2, 200e-6, 4, 500e-6],
                 800e-6, 0.00018593023997294836, [4.00000003, 6.42225962]),
    pytest.param(8, [2, 200e-6, 4, 500e-6],
                 500e-6, 0.0002121557393404887, [3.0000001, 7.42225955]),
    pytest.param(8, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02],
                 500e-6, 0.00047127559879883534, [4.00010831, 134.79617963]),
    pytest.param(1, [1, 0.02, 5, 0.004, 2, 1.23, 1, 0.02],
                 500e-6, 0.0009253340546070372, [-0.89522919, 2.99999999]),
])
def test_calc_distance_for_size(energy, lens_set, fwhm_unf, size_fwhm, expect):
    # old values (with old get_delta):
    # [4.00301939, 134.89793351]
    # [4.00301637, 6.42710261]
    # [3.00226229, 7.42785669]
    # [4.00303623, 134.89791667]
    # [-0.89289053,  2.99216285]
    dis = be_lens_calcs.calc_distance_for_size(size_fwhm=size_fwhm,
                                               lens_set=lens_set,
                                               energy=energy,
                                               fwhm_unfocused=fwhm_unf)
    logger.debug('Expected: %s, Received: %s', expect, dis)
    assert np.allclose(dis, expect)


@pytest.mark.parametrize('energy_sample, distance, expected', [
    pytest.param(8, 4, 4.2643770090253954e-05),
    pytest.param(8, 3, 3.198282756769046e-05),
    pytest.param(9, 4, 3.368672616022848e-05),
    pytest.param(10, 2.5, 1.7051278100008672e-05),
])
def test_find_radius(energy_sample, distance, expected):
    # Old values, with old get_delta
    # 4.261163705976401e-05
    # 3.1958727794823005e-05
    # 3.366056083997648e-05
    # 1.7037713265222187e-05
    radius = be_lens_calcs.find_radius(energy=energy_sample, distance=distance)
    logger.debug('Expected: %s, Received: %s', expected, radius)
    assert np.isclose(expected, radius)


@pytest.mark.parametrize('radius, expected', [
    pytest.param(1.0e-3, 0.0009848857801796104),
    pytest.param(2.0e-3, 0.001392838827718412),
    pytest.param(3.0e-3, 0.0017058722109231982),
])
def test_calc_lens_aperture_radius(radius, expected):
    aperture_radius = be_lens_calcs.calc_lens_aperture_radius(radius)
    logger.debug('Expected: %s, Received: %s', expected,
                 aperture_radius)
    assert np.isclose(expected, aperture_radius)


@pytest.mark.parametrize('energy_sample, radius, expected', [
    pytest.param(8, 1.0e-3, 0.909202465413282),
    pytest.param(8, 2.0e-3, 0.9634630968766913),
    pytest.param(9, 3.0e-3, 0.9806842723309849),
])
def test_calc_trans_for_single_lens(energy_sample, radius, expected):
    # old values (with old get_att_len and old get_delta):
    # 0.9192789600686936
    # 0.9700496829425548
    # 0.9855626484449255
    trans = be_lens_calcs.calc_trans_for_single_lens(energy=energy_sample,
                                                     radius=radius,
                                                     fwhm_unfocused=900e-6)
    logger.debug('Expected: %s, Received: %s', expected, trans)
    assert np.isclose(expected, trans)


@pytest.mark.parametrize('energy_sample, lens_set, expected', [
    pytest.param(8, [2, 100e-6, 4, 200e-6], 0.1909312006707346),
    pytest.param(8, [2, 200e-6, 4, 500e-6], 0.3455637290620128),
    pytest.param(9, [2, 200e-6, 4, 500e-6], 0.3760514142257077),
])
def test_calc_trans_lens_set(energy_sample, lens_set, expected):
    # old values (with old get_att_len and old get_delta):
    # 0.20497965794420414
    # 0.3675986395544651
    # 0.4005461885189993
    trans = be_lens_calcs.calc_trans_lens_set(
        energy_sample, lens_set, fwhm_unfocused=900e-6)
    logger.debug('Expected: %s Received: %s', expected, trans)
    assert np.isclose(expected, trans)


def test_calc_trans_lens_set_with_file_len_set():
    # expected_used_set = (3, 0.0001, 1, 0.0002)
    with patch('pcdscalc.be_lens_calcs.get_lens_set',
               return_value=[3, 0.0001, 1, 0.0002]):
        expected = be_lens_calcs.calc_trans_lens_set(8, [3, 0.0001, 1, 0.0002])

        received = be_lens_calcs.calc_trans_lens_set(8, 1)
        logger.debug('Expected: %s, Received: %s', expected, received)
        assert received == expected


def test_calc_trans_lens_set_with_no_file():
    # make sure the LENS_SET_FILE is none
    # should get an Exception
    be_lens_calcs.LENS_SET_FILE = None
    with pytest.raises(Exception):
        be_lens_calcs.calc_trans_lens_set(8, 1)


@pytest.mark.parametrize('lens_set, distance, expected', [
                         pytest.param([2, 200e-6, 4, 500e-6], 4,
                                      7.0100555419921875),
                         pytest.param([2, 200e-6, 4, 500e-6], 2,
                                      4.9597930908203125),
                         pytest.param([2, 200e-6, 4, 500e-6], 7,
                                      9.270801544189453)
                         ])
def test_find_energy(lens_set, distance, expected):
    # NOTE: the logs from calc_beam_fwhm differ here from the
    # old ones... but the final results seem to be the same???
    # old values (with old get_att_len and old get_delta):
    # 7.007423400878906
    # 4.958213806152344
    # 9.267204284667969
    energy = be_lens_calcs.find_energy(lens_set, distance)
    logger.debug('Expected: %s Received: %s', expected, energy)
    assert np.isclose(expected, energy)


@pytest.mark.parametrize('energy, lens_set, spot_size_fwhm, expected', [
                         pytest.param(8, [2, 200e-6, 4, 500e-6], 4.0e-3,
                                      (-20.844519143534555,
                                      31.266778796088595))
                         ])
def test_find_z_pos(energy, lens_set, spot_size_fwhm, expected):
    # old values (with old get_att_len and old get_delta):
    # (-20.86023779837472, 31.290356777870826)
    z_position = be_lens_calcs.find_z_pos(energy, lens_set, spot_size_fwhm,
                                          fwhm_unfocused=800e-6)
    logger.debug('Expected: %s, Received: %s', expected, z_position)
    assert np.allclose(expected, z_position)


def test_calc_lens_set():
    expected_sets = np.array([[0, 0, 0, 0, 1],
                             [0, 0, 0, 1, 0],
                             [0, 0, 1, 0, 0],
                             [0, 1, 0, 0, 0],
                             [1, 0, 0, 0, 0]])

    expected_effrads = np.array([0.001, 0.0005, 0.0003, 0.0002, 0.0001])
    expected_sizes = np.array([0.00047885, 0.0004574, 0.00042894,
                              0.0003934, 0.00028678])
    expected_foclens = np.array([93.80033687, 46.90016844, 28.14010106,
                                 18.76006737, 9.38003369])

    res = be_lens_calcs.calc_lens_set(energy=8, size_fwhm=2, distance=4,
                                      n_max=1, max_each=2,
                                      lens_radii=[100e-6, 200e-6, 300e-6,
                                                  500e-6, 1000e-6],
                                      fwhm_unfocused=0.0005, eff_rad0=None)

    sets, effrads, sizes, foclens = res

    assert np.allclose(sets, expected_sets)
    assert np.allclose(effrads, expected_effrads)
    assert np.allclose(sizes, expected_sizes)
    assert np.allclose(foclens, expected_foclens)


@pytest.mark.parametrize('radius, fwhm, energy, expected', [
                         pytest.param(2, 200e-6, 8, 0.9937771825941067),
                         pytest.param(4, 500e-6, 9, 0.9953931501494162),
                         pytest.param(6, 300e-6, 10, 0.9964134300156009)
                         ])
def test_lens_transmission(radius, fwhm, energy, expected):
    res = be_lens_calcs.lens_transmission(radius=radius, fwhm=fwhm,
                                          id_material='Be', energy=energy)
    logger.debug('Expected: %s Received: %s', expected, res)
    assert np.isclose(expected, res)


def test_plan_set():
    be_lens_calcs.plan_set(energy=1, z_offset=-10, z_range=[1, 40],
                           beam_size_unfocused=3, size_horizontal=9,
                           size_vertical=None, exclude=[],
                           max_tot_number_of_lenses=1,
                           max_each=5, focus_before_sample=False)

    num, f_m, min_um, max_um, t_percent = be_lens_calcs._plan_set_test_res
    f_m = [round(num, 2) for num in f_m]
    min_um = [round(num, 1) for num in min_um]
    max_um = [round(num, 1) for num in max_um]
    t_percent = [round(num, 1) for num in t_percent]

    expected_num = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    expected_f_m = [0.07, 0.14, 0.28, 0.43, 0.71, 1.42, 2.13, 2.84, 4.27]
    expected_min_um = [466994229.0, 234997114.5, 118998557.3, 80332371.5,
                       49399422.9, 26199711.5, 18466474.3, 14599855.7,
                       10733237.2]
    expected_max_um = [2112064677.4, 1057532338.7,
                       530266169.4, 354510779.6, 213906467.7, 108453233.9,
                       73302155.9, 55726616.9, 38151078.0]

    expected_t_percent = [0, 0, 0, 0, 0, 0, 0, 0, 0]

    assert num == expected_num
    assert np.allclose(f_m, expected_f_m)
    assert np.allclose(min_um, expected_min_um)
    assert np.allclose(max_um, expected_max_um)
    assert np.allclose(t_percent, expected_t_percent)

# Expected based on old code but new get_att_len and get_delta
#  N   f/m   Min/um   Max/um   T/%  Set
#  0  0.07 466994229.0 2112064677.4   0.0  1 x 50um
#  1  0.14 234997114.5 1057532338.7   0.0  1 x 100um
#  2  0.28 118998557.3 530266169.4   0.0  1 x 200um
#  3  0.43 80332371.5 354510779.6   0.0  1 x 300um
#  4  0.71 49399422.9 213906467.7   0.0  1 x 500um
#  5  1.42 26199711.5 108453233.9   0.0  1 x 1000um
#  6  2.13 18466474.3 73302155.9   0.0  1 x 1500um
#  7  2.84 14599855.7 55726616.9   0.0  1 x 2000um
#  8  4.27 10733237.2 38151078.0   0.0  1 x 3000um

# Expected with the old get_att_len and old get_delta
#  N   f/m   Min/um   Max/um   T/%  Set
#  0  0.07 468209535.3 2117588796.8   0.0  1 x 50um
#  1  0.14 235604767.7 1060294398.4   0.0  1 x 100um
#  2  0.28 119302383.8 531647199.2   0.0  1 x 200um
#  3  0.43 80534922.6 355431466.1   0.0  1 x 300um
#  4  0.71 49520953.5 214458879.7   0.0  1 x 500um
#  5  1.42 26260476.8 108729439.8   0.0  1 x 1000um
#  6  2.13 18506984.5 73486293.2   0.0  1 x 1500um
#  7  2.84 14630238.4 55864719.9   0.0  1 x 2000um
#  8  4.26 10753492.3 38243146.6   0.0  1 x 3000um
# Got
# N   f/m   Min/um   Max/um   T/% Set
# 0  0.07 466994229.0 2112064677.4   0.0  1 x 50um
# 1  0.14 234997114.5 1057532338.7   0.0  1 x 100um
# 2  0.28 118998557.3 530266169.4   0.0  1 x 200um
# 3  0.43 80332371.5 354510779.6   0.0  1 x 300um
# 4  0.71 49399422.9 213906467.7   0.0  1 x 500um
# 5  1.42 26199711.5 108453233.9   0.0  1 x 1000um
# 6  2.13 18466474.3 73302155.9   0.0  1 x 1500um
# 7  2.84 14599855.7 55726616.9   0.0  1 x 2000um
#  8  4.27 10733237.2 38151078.0   0.0  1 x 3000um



"""
Module that handles calculations for diffraction/crystallography.

Assume all the calculations that expect energy as input will
require energy in eV.
"""
import numpy as np

from .common import asind, cosd, energy_to_wavelength, sind
from .constants import chemical_name_to_formula, lattice_parameters, units


def get_lom_geometry(energy, material_id, reflection):
    """
    Calculate the Large Offset Monochromator's crystal geometry.

    Parameters
    ----------
    energy : number
        Photon energy in eV.
    material_id : str
        Chemical formula. E.g.: `Si`.
    reflection : tuple
        The reflection. E.g.: `(1,1,1)`.

    Returns
    -------
    thm, zm : tuple
    """
    th = np.radians(bragg_angle(material_id, reflection, energy))
    zm = 300 / np.tan(2 * th)
    thm = np.rad2deg(th)
    return thm, zm


def bragg_angle(material_id, hkl, energy):
    """
    Compute the Bragg angle in deg.

    Computes the bragg angle of the specified material, reflection and photon
    energy.

    Parameters
    ----------
    material_id : str
        Chemical formula. Defaults to `Si`.
    hkl : tuple
        The reflection indices. Defaults to `(1,1,1)`.
    energy : number
        The photon energy in eV.

    Returns
    -------
    theta : number
        Theta in degrees.
    """
    material_id = chemical_name_to_formula.get(material_id, material_id)
    d = d_space(material_id, hkl)
    theta = asind(energy_to_wavelength(energy) / 2 / d)
    return theta


def d_space(material_id, hkl):
    """
    Compute the d spacing (m) of the specified material and reflection.

    The d-spacing can be described as the distance between planes of atoms
    that give rise to diffraction peaks. Each peak in a diffractogram results
    from a corresponding d-spacing. The planes of atoms can be referred to a
    3D coordinate system and so can be described as a direction within the
    crystal. So d-spacing as well as having a dimension, usually quoted in
    Ångstroms, can be labelled with a plane direction hkl.

    Parameters
    ----------
    material_id : str
        Chemical fomula. E.g.: `Si`
    hkl : tuple
        Miller Indices, the reflection. E.g.: `(1,1,1)`

    Returns
    -------
    d : number
        The lattice plane spacing d in Bragg's Law for the chosen plane
        indicated by the Miller indices.
    """
    material_id = chemical_name_to_formula.get(material_id, material_id)
    h_index, k_index, l_index = hkl
    lp = lattice_parameters[material_id]
    # a, b, c in angstroms
    # alpha, beta, gamma in degrees
    a, b, c, alpha, beta, gamma = lp
    a = a / units["ang"]
    b = b / units["ang"]
    c = c / units["ang"]

    cos_alpha = cosd(alpha)
    cos_beta = cosd(beta)
    cos_gamma = cosd(gamma)
    sin_alpha = sind(alpha)
    sin_beta = sind(beta)
    sin_gamma = sind(gamma)

    inv_d_sqr = (
        1 / (1 + 2 * cos_alpha * cos_beta * cos_gamma - cos_alpha ** 2
             - cos_beta ** 2 - cos_gamma ** 2)
        * (
            h_index ** 2 * sin_alpha ** 2 / a ** 2
            + k_index ** 2 * sin_beta ** 2 / b ** 2
            + l_index ** 2 * sin_gamma ** 2 / c ** 2
            + 2 * h_index * k_index *
            (cos_alpha * cos_beta - cos_gamma) / a / b
            + 2 * k_index * l_index *
            (cos_beta * cos_gamma - cos_alpha) / b / c
            + 2 * h_index * l_index *
            (cos_alpha * cos_gamma - cos_beta) / a / c
        )
    )
    d = inv_d_sqr ** -0.5
    return d



import logging

import pytest

from pcdscalc.pmps import (KFE, LFE, check_actual_range, check_bitmask,
                           describe_bitmask, get_bitmask,
                           select_bitmask_boundaries)

logger = logging.getLogger(__name__)

# 32 bits, using numbers from 1 to 32
test_boundaries = list(range(1, 33))

# Define some utility bitmasks
allow_none = 0
allow_all = 2**32-1
bm1 = 0b1111_0000_0000_0000_0000_0000_0000_1111
bm2 = 0b0000_0000_0000_1111_1111_0000_0000_0000
bm3 = 0b0000_0000_0000_0001_0000_0000_0000_0000
bm4 = 0b1111_1111_1111_1110_1111_1111_1111_1111


@pytest.mark.parametrize(
    "test_input,expected",
    [('k', KFE), ('kfe', KFE), ('sxr', KFE),
     ('L', LFE), ('LFE', LFE), ('HXR', LFE)]
)
def test_select_bounds(test_input, expected):
    logger.debug(f'test_select_bounds({test_input}, {expected})')
    assert select_bitmask_boundaries(test_input) is expected


@pytest.mark.parametrize(
    "lower,upper,allow,expected",
    [(0, 100, True, allow_all),
     (0, 100, False, allow_none),
     (0, 15.5, True,     0b0000_0000_0000_0000_0111_1111_1111_1111),  # noqa: E241
     (15.5, 100, True,   0b1111_1111_1111_1111_0000_0000_0000_0000),  # noqa: E241
     (14.5, 21.5, True,  0b0000_0000_0001_1111_1000_0000_0000_0000),  # noqa: E241
     (14.5, 21.5, False, 0b1111_1111_1100_0000_0011_1111_1111_1111),  # noqa: E241
     (15, 20, True,      0b0000_0000_0000_1111_1000_0000_0000_0000),  # noqa: E241
     (15, 20, False,     0b1111_1111_1111_0000_0111_1111_1111_1111),  # noqa: E241
     ]
)
def test_get_bitmask(lower, upper, allow, expected):
    """
    Test that the correct bitmask is created.

    Explanation of test cases 3 to 8 (first two are obvious):
    3. Allow between 0 and 15, exclude the 15.5 point because we can't allow
       points like 15.6 and they share a range. Therefore, enable bits 1
       through 15 (bit 1 allows 0 to 1, bit 15 allows 14 to 15).
    4. Allow between 16 and 32, exclude the 15.5 point because we can't allow
       points like 15.4 and they share a range. Cut off at 32 for the top of
       the bitmask range. Therefore, enable bits 17 through 32 (bit 17 allows
       16 to 17).
    5. Allow between 15 and 21, exclude the exact points because we can't allow
       any points outside the range. Enable bits 16 to 21 (bit 16 allows 15 to
       16, bit 21 allows 20 to 21).
    6. Allow between 0 and 14 and between 22 and 32. Turn off bits 15 to 22
       (bit 15 allows 14 to 15, bit 22 allows 21 to 22). Exclude the boundary
       bits for the same reasonings as above.
    7. Allow exactly between 15 and 20. Turn on bits 16 to 20 (bit 16 allows 15
       to 16, bit 20 allows 19 to 20)
    8. Turn off bits 16 to 20 (bit 16 allows 15 to 16, bit 20 allows 19 to 20)
    """
    logger.debug(f'test_get_bitmask({lower}, {upper}, {allow}, {expected})')
    bitmask = get_bitmask(lower, upper, allow, 'tst', bounds=test_boundaries)
    assert bitmask == expected


@pytest.mark.parametrize(
    "energy,bitmask,expected",
    [(-1, allow_all, False), (100, allow_all, False), (16, allow_all, True),
     (0, allow_none, False), (15, allow_none, False), (30, allow_none, False),
     (0, bm1, True), (16, bm1, False), (30, bm1, True), (40, bm1, False),
     (0, bm2, False), (16, bm2, True), (30, bm2, False), (-1, bm2, False),
     (7, bm3, False), (16, bm3, False), (16.5, bm3, True), (17, bm3, False),
     (15, bm4, True), (16, bm4, False), (16.5, bm4, False), (17, bm4, False)])
def test_check_bitmask(energy, bitmask, expected):
    logger.debug(f'test_check_bitmask({energy}, {bitmask}, {expected}')
    ok = check_bitmask(energy, bitmask, 'tst', bounds=test_boundaries)
    assert ok == expected


@pytest.mark.parametrize(
    "lower,upper,allow,expected",
    [(0, 100, True, (0, 32)),
     (10, 20, True, (10, 20)),
     (10.5, 20.5, True, (11, 20)),
     (11, 21, False, (11, 21)),
     (9.5, 15.5, False, (9, 16)),
     (10.4, 10.6, True, (10.4, 10.4))])
def test_actual_range(lower, upper, allow, expected):
    logger.debug(f'test_actual_range({lower}, {upper}, {allow}, {expected})')
    span = check_actual_range(lower, upper, allow, 'tst',
                              bounds=test_boundaries)
    assert span == expected


@pytest.mark.parametrize(
    "test_input",
    [allow_none, allow_all, bm1, bm2, bm3, bm4])
def test_describe_bitmask(test_input):
    logger.debug(f'test_describe_bitmask({test_input})')
    describe_bitmask(test_input, 'tst', bounds=test_boundaries)



"""Module that holds common calculations."""
import numpy as np

from .constants import WAVELENGTH_TO_ENERGY_LAMBDA, units


def cosd(angle):
    """
    Cos of an angle specified in degrees.

    Parameters
    ----------
    angle : number
        Angle in degrees.

    Returns
    -------
    arad : number
        The corresponding cosine value of `angle` input.
    """
    arad = np.deg2rad(angle)
    return np.cos(arad)


def sind(angle):
    """
    Sin of an angle specified in degrees.

    Parameters
    ----------
    angle : number
        Angle in degrees.

    Returns
    -------
    arad : number
        The corresponding sine value of `angle` input.
    """
    arad = np.deg2rad(angle)
    return np.sin(arad)


def tand(angle):
    """
    Tan of an angle specified in degrees.

    Parameters
    ----------
    angle : number
        Angle in degrees.

    Returns
    -------
    arad : number
        The corresponding tangent value of `angle` input.
    """
    arad = np.deg2rad(angle)
    return np.tan(arad)


def asind(x):
    """
    Arcsin in degrees. Closed interval of [-pi/2, pi/2].

    Parameters
    ----------
    x : number
        y-coordinate on the unit circle.

    Returns
    -------
    arad : number
        The inverse sine of x in degrees.
    """
    angle = np.arcsin(x)
    return np.rad2deg(angle)


def acosd(x):
    """
    Arccos in degrees.

    Parameters
    ----------
    x : number
        x-coordinate on the unit circle.
        For real arguments the domain is [-1, 1].

    Returns
    -------
    arad : number
        The angle of the ray intersecting the unit circle at the given
        x-coordinate in degrees.
    """
    angle = np.arccos(x)
    return np.rad2deg(angle)


def atand(x):
    """
    Arctan in degrees.

    Parameters
    ----------
    x : number

    Returns
    -------
    arad : number
        Has the same shape as x. Its real part is in [-pi/2, pi/2]
        (arctan(+/-inf) returns +/-pi/2).
    """
    angle = np.arctan(x)
    return np.rad2deg(angle)


def energy_to_wavelength(energy):
    """
    Compute photon wavelength in m.

    Parameters
    ----------
    energy : number
        Photon energy in eV.

    Returns
    -------
    wavelength : float
        Wavelength [m].
    """
    return (WAVELENGTH_TO_ENERGY_LAMBDA / energy) / units["ang"]


def wavelength_to_energy(wavelength):
    """
    Compute photon energy in eV.

    Parameters
    ----------
    wavelength : number
        The photon wavelength in m.

    Returns
    -------
    energy : number
        Photon Energy in eV.
    """
    return WAVELENGTH_TO_ENERGY_LAMBDA / (wavelength * units["ang"])
